#!/usr/bin/env python
# obs.py

"""
Author : peters 

Revision History:
File created on 28 Jul 2010.

"""
import os
import sys
import logging
import datetime

identifier = 'CarbonTracker CO2 mixing ratios'
version = '0.0'

from da.baseclasses.obs import Observation

################### Begin Class CtObservations ###################

class CtObservations(Observation):
    """ an object that holds data + methods and attributes needed to manipulate mixing ratio values """

    def getid(self):
        return identifier

    def getversion(self):
        return identifier

    def Initialize(self,DaCycle):

        self.startdate = DaCycle.da_settings['time.sample.start']
        self.enddate   = DaCycle.da_settings['time.sample.end']
        DaSystem       = DaCycle.DaSystem

        sfname      = DaSystem.da_settings['obs.input.fname']
        filename    = os.path.join(DaSystem.da_settings['obs.input.dir'],sfname)

        if not os.path.exists(filename):
            msg     = 'Could not find  the required observation input file (%s) ' % filename ; logging.error(msg)
            raise IOError,msg
        else:
            self.ObsFilename = filename

        self.Data = MixingRatioList([])

        return None

    def AddObs(self):
        """ Returns a MixingRatioList holding individual MixingRatioSample objects for all obs in a file
      
            The CarbonTracker mixing ratio files are provided as one long list of obs for all possible dates. So we can 
            either:
            
            (1) read all, and the subselect the data we will use in the rest of this cycle
            (2) Use nco to make a subset of the data
            
            For now, we will stick with option (1) 
        
        """
        import Nio
        import datetime as dtm
        from string import strip, join
        from numpy import array, logical_and

        ncf         = Nio.open_file(self.ObsFilename)
        idates      = ncf.variables['date_components'].get_value()
        dates       = array([dtm.datetime(*d) for d in idates])

        subselect   = logical_and(dates >= self.startdate , dates <= self.enddate).nonzero()[0]

        dates       = dates.take(subselect,axis=0)
        
        ids         = ncf.variables['id'].get_value().take(subselect,axis=0)
        sites       = ncf.variables['site'].get_value().take(subselect,axis=0)
        sites       = [s.tostring().lower() for s in sites]
        sites       = map(strip,sites)
        lats        = ncf.variables['lat'].get_value().take(subselect,axis=0)
        lons        = ncf.variables['lon'].get_value().take(subselect,axis=0)
        alts        = ncf.variables['alt'].get_value().take(subselect,axis=0)
        obs         = ncf.variables['obs'].get_value().take(subselect,axis=0)
        species     = ncf.variables['species'].get_value().take(subselect,axis=0)
        date        = ncf.variables['date'].get_value().take(subselect,axis=0)
        window      = ncf.variables['sampling_strategy'].get_value().take(subselect,axis=0)
        flags       = ncf.variables['NOAA_QC_flags'].get_value().take(subselect,axis=0)
        flags       = [s.tostring().lower() for s in flags]
        flags       = map(strip,flags)
        dummy       = ncf.close()

        msg         = "Successfully read data from obs file (%s)"%self.ObsFilename ; logging.debug(msg)

       
        for n in range(len(dates)): 

           self.Data.append( MixingRatioSample(dates[n],sites[n],obs[n],0.0,0.0,0.0,0.0,flags[n],alts[n],lats[n],lons[n],ids[n],0.0) )

        msg         = "Added %d observations to the Data list" % len(dates) ; logging.debug(msg)

    def AddSimulations(self,filename,silent=True):
        """ Adds model simulated values to the mixing ratio objects """

        import Nio

        ncf         = Nio.open_file(filename,format='hdf')
        ids         = ncf.variables['id'].get_value()
        simulated   = ncf.variables['sampled_mole_frac'].get_value()*1e6
        dummy       = ncf.close()
        msg         = "Successfully read data from model sample file (%s)"%filename ; logging.info(msg)

        obs_ids     = self.Data.getvalues('id')

        obs_ids     = obs_ids.tolist()
        ids         = map(int,ids)

        missing_samples = []

        for id,val in zip(ids,simulated): 
           
            if id in obs_ids:
                
                index = obs_ids.index(id)
                dummy = self.Data[index].simulated = val

            else:     

                missing_samples.append(id)

        if not silent and missing_samples != []:
            msg = 'Model samples were found that did not match any ID in the observation list. Skipping them...' ; logging.warning(msg)
            msg = '%s'%missing_samples ; logging.warning(msg)

        msg  = "Added %d simulated values to the Data list" % (len(ids)-len(missing_samples) ) ; logging.debug(msg)

    def WriteSampleInfo(self, DaCycle):
        """ 
            Write the information needed by the observation operator to a file. Return the filename that was written for later use

            For CarbonTracker, the input files are already in a format suitable for TM5 to read. All we need to do return the name of the file 
            that served as input for the Observation object.

        """
        import shutil

        obsinputfile   = os.path.join(DaCycle.da_settings['dir.input'],'observations.nc')

        dummy          = shutil.copy(self.ObsFilename,obsinputfile)
        msg         = "           [copying source] .... %s " % self.ObsFilename  ; logging.debug(msg)
        msg         = "           [to destination] .... %s " % obsinputfile      ; logging.debug(msg)
        msg         = "Sample input file for obs operator now in place (%s)"%obsinputfile ; logging.info(msg)

        return obsinputfile



################### End Class CtObservations ###################



################### Begin Class MixingRatioSample ###################

class MixingRatioSample(object):
    """ 
        Holds the data that defines a Mixing Ratio Sample in the data assimilation framework. Sor far, this includes all
        attributes listed below in the __init__ method. One can additionally make more types of data, or make new
        objects for specific projects.

    """

    def __init__(self,xdate,code='XXX',obs=0.0,simulated=0.0,resid=0.0,hphr=0.0,mdm=0.0,flag=0,height=0.0,lat=-999.,lon=-999.,evn='0000',sdev=0.0):

            self.code       = code.strip()      # Site code
            self.xdate      = xdate             # Date of obs
            self.obs        = obs               # Value observed
            self.simulated  = simulated         # Value simulated by model
            self.resid      = resid             # Mixing ratio residuals
            self.hphr       = hphr              # Mixing ratio prior uncertainty from fluxes and (HPH) and model data mismatch (R)
            self.mdm        = mdm               # Model data mismatch
            self.flag       = flag              # Flag
            self.height     = height            # Sample height
            self.lat        = lat               # Sample lat
            self.lon        = lon               # Sample lon
            self.id         = evn               # Event number
            self.sdev       = sdev              # standard deviation of ensemble
            self.masl       = True              # Sample is in Meters Above Sea Level
            self.mag        = not self.masl     # Sample is in Meters Above Ground

    def __str__(self):
            day=self.xdate.strftime('%Y-%m-%d %H:%M:%S')
            return ' '.join(map(str,[self.code,day,self.obs,self.flag,self.id]))

################### End Class MixingRatioSample ###################

################### Begin Class MixingRatioList ###################

class MixingRatioList(list):
    """ This is a special type of list that holds MixingRatioSample objects. It has methods to extract data from such a list easily """
    from numpy import array, ndarray

    def getvalues(self,name,constructor=array):
            from numpy import ndarray
            result = constructor([getattr(o,name) for o in self])
            if isinstance(result,ndarray): 
                return result.squeeze()
            else:
                return result
    def unflagged(self):
            return MixingRatioSample([o for o in self if o.flag == 0])
    def flagged(self):
            return MixingRatioSample([o for o in self if o.flag != 0])
    def selectsite(self,site='mlo'):
            l=[o for o in self if o.code == site]
            return MixingRatioSample(l)
    def selecthours(self,hours=range(1,24)):
            l=[o for o in self if o.xdate.hour in hours]
            return MixingRatioSample(l)

################### End Class MixingRatioList ###################

