#!/usr/bin/env python
# pipeline.py

"""
Author : peters 

Revision History:
File created on 26 Aug 2010.

"""
import logging
import os
import sys
import shutil
import datetime
import da.tools.rc as rc

header  = '\n\n    ***************************************   '
footer  = '    *************************************** \n  '

validprocess=['jobstart','jobinput','sample','invert','propagate','resubmit','all']

system_not_implemented_message = "Unable to import a system specific class or function, exiting..."

def Main(Samples,StateVector,ObsOperator,Optimizer):
    """ The main point of entry for the pipeline """

# Append current working dir to path

    sys.path.append(os.getcwd())

# Import methods and classes contained in this package

    from da.tools.initexit import CleanUpCycle
    from da.tools.initexit import ValidateOptsArgs
    from da.tools.initexit import ParseOptions
    from da.tools.pipeline import JobStart, JobInput, SampleState, Invert, Advance, SaveAndSubmit 
    from da.tools.general import StoreData,RestoreData
    

# Parse options from the command line

    opts, args = ParseOptions()

# Validate Options and arguments passed

    opts,args   = ValidateOptsArgs(opts,args)

# Create the Cycle Control object for this job    

    DaCycle      = JobStart(opts,args)

    msg          = header+"Initializing Da Cycle"+footer           ; logging.info(msg) 

    dummy        = DaCycle.Initialize()

    msg          = header+"starting JobInput"+footer           ; logging.info(msg) 

    StateVector  = JobInput(DaCycle, StateVector)

    msg          = header+"starting SampleState"+footer             ; logging.info(msg) 

    dummy        = SampleState(DaCycle,Samples,StateVector,ObsOperator)

    msg          = header+"starting Invert"+footer             ; logging.info(msg) 

    dummy        = Invert(DaCycle, StateVector, Optimizer)

    msg          = header+"starting Advance"+footer          ; logging.info(msg) 

    dummy        = Advance(DaCycle, Samples, StateVector, ObsOperator)

    msg          = header+"starting SaveAndSubmit"+footer      ; logging.info(msg) 

    dummy        = SaveAndSubmit(DaCycle, StateVector)

    msg          = "Cycle finished...exiting"     ; logging.info(msg)

    dummy        = CleanUpCycle(DaCycle)

    sys.exit(0)


####################################################################################################

def JobStart(opts,args):
    """ Set up the job specific directory structure and create an expanded rc-file """
    from da.tools.initexit import CycleControl

    # First create a CycleControl object that handles all the details of the da cycle

    DaCycle     = CycleControl(opts,args)

    # Next create a DaSystem object that contains all the details related to the DA system

    if DaCycle.da_settings['da.system'] == 'CarbonTracker':
        from da.ct.dasystem import CtDaSystem as DaSystem
    else:
        from da.baseclasses.dasystem import DaSystem 

    DaSystem    = DaSystem(DaCycle.da_settings['da.system.rc'])

    dummy       = DaSystem.Validate()

    DaCycle.DaSystem = DaSystem

    # Next, get the information on the platform

    if DaCycle.da_settings['da.platform'] == 'jet':
        from da.platform.jet import JetPlatForm as PlatForm
    elif DaCycle.da_settings['da.platform'] == 'maunaloa':
        from da.platform.maunaloa import MaunaloaPlatForm as PlatForm
    else:
        from da.baseclasses.platform import PlatForm

    DaPlatForm  = PlatForm()

    DaCycle.DaPlatForm = DaPlatForm

    return DaCycle


def JobInput(DaCycle, StateVector):
    """ Set up the input data for the forward model: obs and parameters/fluxes"""

    dims        = ( int(DaCycle.da_settings['time.nlag']),
                    int(DaCycle.da_settings['forecast.nmembers']),
                    int(DaCycle.DaSystem.da_settings['nparameters']),
                  )

    dummy       = StateVector.Initialize(dims)

    nlag        = dims[0]

    # We now have an empty StateVector object that we need to populate with data. If this is a continuation from a previous cycle, we can read
    # the previous StateVector values from a NetCDF file in the save/ directory. If this is the first cycle, we need to populate the StateVector
    # with new values for each week. After we have constructed the StateVector, it will be propagated by one cycle length so it is ready to be used
    # in the current cycle

    if not DaCycle.da_settings['time.restart']:

        # Fill each week from n=1 to n=nlag with a new ensemble

        for n in range(0,nlag):
            cov   = StateVector.GetCovariance(DaCycle.DaSystem)
            dummy = StateVector.MakeNewEnsemble(n+1,cov)

    else:

        # Read the StateVector data from file

        savedir         = DaCycle.da_settings['dir.save']
        filtertime      = DaCycle.da_settings['time.start'].strftime('%Y%m%d')
        filename        = os.path.join(savedir,'savestate.nc')

        StateVector.ReadFromFile(filename)

        # Now propagate the ensemble by one cycle to prepare for the current cycle

        dummy = StateVector.Propagate()

    return StateVector

def SampleState(DaCycle,Samples,StateVector, ObservationOperator):
    """ Sample the filter state for the inversion """

    from da.tools.general import AdvanceTime

    # Before a forecast step, save all the data to a save/tmp directory so we can later recover it before the propagation step.
    # This is especially important for:
    #  (i) The transport model restart data which holds the background mixing ratios. This is needed to run the model one cycle forward
    #  (ii) The random numbers (or the seed for the random number generator) so we can recreate the ensembles if needed

    dummy   = DaCycle.MoveSaveData(io_option='store',save_option='partial',filter=[])
    msg     = "All restart data have been copied to the save/tmp directory for future use"    ; logging.debug(msg)

    nlag    = int(DaCycle.da_settings['time.nlag'])
    msg     = "Sampling model will be run over %d cycles"% nlag ; logging.info(msg)

    for lag in range(nlag):

        msg     = header+".....Ensemble Kalman Filter at lag %d"% (int(lag)+1,)           ; logging.info(msg) 

        # First set up the information for time start and time end of this sample

        cyclelen    = DaCycle.da_settings['cyclelength']

        if lag == 0:
            startdate   = DaCycle.da_settings['time.start']
        else:
            startdate   = DaCycle.da_settings['time.sample.start']  


        enddate = AdvanceTime(startdate,cyclelen)

        DaCycle.da_settings['time.sample.start']  = startdate
        DaCycle.da_settings['time.sample.end']    = enddate
        DaCycle.da_settings['time.sample.window'] = lag

        msg   = "New simulation interval set : "                                       ; logging.info(msg)
        msg   = "                  start date : %s " % startdate.strftime('%F %H:%M')  ; logging.info(msg)
        msg   = "                  end   date : %s " % enddate.strftime('%F %H:%M')    ; logging.info(msg)


        # Implement something that writes the ensemble member parameter info to file, or manipulates them further into the 
        # type of info needed in your transport model

        dummy   = WriteEnsembleData(DaCycle, StateVector, lag)

        dummy   = Samples.Initialize(DaCycle)

        dummy   = Samples.Validate() 

        dummy   = Samples.AddObs() 

        filename   = Samples.WriteSampleInfo(DaCycle) 

        DaCycle.da_settings['ObsOperator.inputfile'] = filename

        # Run the forecast model

        dummy   = RunForecastModel(DaCycle,ObservationOperator)

        # Read forecast model samples that were written to NetCDF files by each member, also add the obs to the statevector

        silent = False
        for Member in StateVector.EnsembleMembers[lag]:
            filename               = os.path.join(DaCycle.da_settings['dir.output'],'samples.%03d.nc'%Member.membernumber)
            dummy                  = Samples.AddSimulations(filename,silent)
            Member.ModelSample     = Samples
            silent=True

        msg = "Added samples from the forecast model to each member of the StateVector" ; logging.debug(msg)

        # Advance the sample time

        startdate = AdvanceTime( startdate, cyclelen)
        DaCycle.da_settings['time.sample.start']  = startdate

    # Optionally, post-processing of the model output can be added that deals for instance with
    # sub-sampling of time series, vertical averaging, etc.

    return None

def Invert(DaCycle, StateVector, Optimizer ):
    """ Perform the inverse calculation """

    nobs      = len(StateVector.EnsembleMembers[-1][0].ModelSample.Data.getvalues('obs'))

    dims      = ( int(DaCycle.da_settings['time.nlag']),
                  int(DaCycle.da_settings['forecast.nmembers']),
                  int(DaCycle.DaSystem.da_settings['nparameters']),
                  nobs,  )

    dummy     = Optimizer.Initialize(dims)

    dummy     = Optimizer.StateToMatrix(StateVector )

    dummy     = Optimizer.SetLocalization('None')

    if not DaCycle.DaSystem.da_settings.has_key('opt.algorithm'):
        msg       = "There was no minimum least squares algorithm specified in the DA System rc file (key : opt.algorithm)" ; logging.info(msg) 
        msg       = "...using serial algorithm as default..." ; logging.info(msg)
        dummy     = Optimizer.SerialMinimumLeastSquares()
    elif DaCycle.DaSystem.da_settings['opt.algorithm'] == 'serial':
        msg       = "Using the serial minimum least squares algorithm to solve ENKF equations" ; logging.info(msg)
        dummy     = Optimizer.SerialMinimumLeastSquares()
    elif DaCycle.DaSystem.da_settings['opt.algorithm'] == 'bulk':
        msg       = "Using the bulk minimum least squares algorithm to solve ENKF equations" ; logging.info(msg)
        dummy     = Optimizer.BulkMinimumLeastSquares()


    dummy     = Optimizer.MatrixToState(StateVector )

    StateVector.isOptimized = True

    return None

def Advance( DaCycle, Samples, StateVector, ObservationOperator):
    """ Advance the filter state to the next step """
    from da.tools.pipeline import WriteEnsembleData
    from da.tools.pipeline import RunForecastModel
    from da.tools.general import AdvanceTime

    # This is the advance of the modeled CO2 state. Optionally, routines can be added to advance the state vector (mean+covariance)

    # Then, restore model state from the start of the filter

    dummy   = DaCycle.MoveSaveData(io_option='restore',save_option='partial',filter=[])
    msg     = "All restart data have been recovered from the save/tmp directory, this "+ \
              "resets the filter to %s "%DaCycle.da_settings['time.start']   ; logging.debug(msg)

    msg     = "Sampling model will be run over 1 cycle"  ; logging.info(msg)

    # First, get new sampling dates to collect the final model samples

    # Get the observations corresponding to the current cycle

    cyclelen    = DaCycle.da_settings['cyclelength']

    startdate   = DaCycle.da_settings['time.start']
    enddate = AdvanceTime(startdate,cyclelen)

    DaCycle.da_settings['time.sample.start']  = startdate
    DaCycle.da_settings['time.sample.end']    = enddate
    DaCycle.da_settings['time.sample.window'] = 0

    msg   = "New simulation interval set : "                                       ; logging.info(msg)
    msg   = "                  start date : %s " % startdate.strftime('%F %H:%M')  ; logging.info(msg)
    msg   = "                  end   date : %s " % enddate.strftime('%F %H:%M')    ; logging.info(msg)

    # Start the logic of the cycle, first get a list of observations to sample

    dummy      = Samples.Initialize(DaCycle)

    dummy      = Samples.Validate() 

    dummy      = Samples.AddObs() 

    # Write a file that serves as input for the observation operator

    filename   = Samples.WriteSampleInfo(DaCycle) 

    DaCycle.da_settings['ObsOperator.inputfile'] = filename

    # Write the new ensembles from the statevector to disk, for the cycle that is now finished

    dummy   = WriteEnsembleData(DaCycle, StateVector, 0)

    # Run the forecast model once more to advance the background CO2 state, and to sample the model once more

    dummy   = RunForecastModel(DaCycle,ObservationOperator)

    # Read forecast model samples that were written to NetCDF files by each member, also add the obs to the statevector

    silent = False
    for Member in StateVector.EnsembleMembers[0]:
        filename               = os.path.join(DaCycle.da_settings['dir.output'],'samples.%03d.nc'%Member.membernumber)
        dummy                  = Samples.AddSimulations(filename,silent)
        Member.ModelSample     = Samples
        silent=True

    msg = "Added samples from the forecast model to each member of the StateVector" ; logging.debug(msg)

    return None

def SaveAndSubmit( DaCycle, StateVector):
    """ Save the model state and submit the next job """

    savedir         = DaCycle.da_settings['dir.output']
    filename        = os.path.join(savedir,'savestate.nc')

    dummy = StateVector.WriteToFile(filename)
    dummy = DaCycle.Finalize()

    return None

def RunForecastModel(DaCycle,ObsOperator):
    """Prepare and execute a forecast step using an external Fortran model. Note that the flavor of model 
       used is determined in the very first line where the import statement of module "model" depends on a 
       setting in your da.rc file. After that choice, the module written specifically for a particular 
       application takes over. There are a few required variables and methods in such a module:
       
       version [variable, string]    : defines the version number of the module, e.g. "1.0"
       identifier [variable, string] : identifies the model used as a string variable, e.g. "TM5"
       PrepareExe (method)           : A method that preps the model simulation. It can for instance create an input 
                                       parameter file for the model (tm5.rc), or execute some scripts needed before the 
                                       actual execution starts (get_meteo). After this step, a call to the model 
                                       executable should start a succesfull simulation
       Run        (method)           : Start the executable. How this is done depends on your model and might involve 
                                       submitting a string of jobs to the queue, or simply spawning a subprocess, or ...
       """

    status = ObsOperator.Initialize(DaCycle)

    status = ObsOperator.ValidateInput(DaCycle)

    #status = ObsOperator.Run(DaCycle)

    dummy = ObsOperator.SaveData()

    return status

def WriteEnsembleData(DaCycle, StateVector, lag):
    """ Write the data needed by each model ensemble member to disk"""

    members =  StateVector.EnsembleMembers[lag]

    for mem in members:

        mem.WriteToFile(DaCycle.da_settings['dir.input'])

    msg = "Ensemble member parameter files were written to disk for lag = %d" % lag ; logging.info(msg)
    return None


